#ifndef PKMNLIB_ANGELSCRIPTTYPEINFO_HPP
#define PKMNLIB_ANGELSCRIPTTYPEINFO_HPP

#define ANGELSCRIPT_DLL_LIBRARY_IMPORT
#include <Arbutils/Collections/Dictionary.hpp>
#include <Arbutils/ConstString.hpp>
#include <CreatureLib/Library/Exceptions/CreatureException.hpp>
#include <angelscript.h>
#include <cstring>
#include <unordered_map>
#include <utility>

using ConstString = Arbutils::CaseInsensitiveConstString;

class AngelScriptTypeInfo {
private:
    asITypeInfo* _type = nullptr;
    Arbutils::Collections::Dictionary<uint32_t, asIScriptFunction*> _functions;
    ConstString _name;

    struct FunctionInfo {
        bool Exists = false;
        asIScriptFunction* Function = nullptr;
    };

    FunctionInfo Initialize(const std::string& decl) {
        auto val = _type->GetMethodByDecl(decl.c_str(), false);
        if (val == nullptr) {
            return FunctionInfo{.Exists = false, .Function = nullptr};
        }
        if (!val->IsOverride()) {
            return FunctionInfo{.Exists = false, .Function = nullptr};
        }
        return FunctionInfo{.Exists = true, .Function = val};
    }

public:
    explicit AngelScriptTypeInfo(const ConstString& name, asITypeInfo* type) : _type(type), _name(name) {}
    ~AngelScriptTypeInfo() {
        for (const auto& f : _functions) {
            f.second->Release();
        }
        _functions.Clear();
    }

    const ConstString& GetName() const noexcept { return _name; }

    const char* GetDecl() { return _type->GetName(); }

    asIScriptFunction* GetFunction(const ConstString& functionName) {
        asIScriptFunction* func;
        if (_functions.TryGet(functionName, func)) {
            return func;
        }
        func = _type->GetMethodByName(functionName.c_str());
        if (func != nullptr) {
            func->AddRef();
        }
        _functions.Insert(functionName, func);
        return func;
    }

    asIScriptObject* Instantiate(asIScriptContext* ctx) {
        auto factory = _type->GetFactoryByIndex(0);
        ctx->Prepare(factory);
        auto result = ctx->Execute();
        if (result != asEXECUTION_FINISHED) {
            throw CreatureException("Instantiation failed.");
        }
        asIScriptObject* obj = *(asIScriptObject**)ctx->GetAddressOfReturnValue();
        obj->AddRef();
        return obj;
    }

#define SCRIPT_HOOK_FUNCTION(name, decl)                                                                               \
private:                                                                                                               \
    FunctionInfo __##name = Initialize(decl);                                                                          \
                                                                                                                       \
public:                                                                                                                \
    const FunctionInfo& Get##name() const { return __##name; }

    SCRIPT_HOOK_FUNCTION(OnInitialize, "void OnInitialize(const array<EffectParameter@> &in parameters)");
    SCRIPT_HOOK_FUNCTION(Stack, "void Stack()");
    SCRIPT_HOOK_FUNCTION(OnRemove, "void OnRemove()");
    SCRIPT_HOOK_FUNCTION(PreventAttack, "void PreventAttack(ExecutingMove@ attack, bool& result)");
    SCRIPT_HOOK_FUNCTION(FailAttack, "void FailAttack(ExecutingMove@ attack, bool& result)");
    SCRIPT_HOOK_FUNCTION(StopBeforeAttack, "void StopBeforeAttack(ExecutingMove@ attack, bool& result)");
    SCRIPT_HOOK_FUNCTION(OnBeforeAttack, "void OnBeforeAttack(ExecutingMove@ attack)");
    SCRIPT_HOOK_FUNCTION(FailIncomingAttack,
                         "void FailIncomingAttack(ExecutingMove@ attack, Pokemon@ target, bool& result)");
    SCRIPT_HOOK_FUNCTION(IsInvulnerable, "void IsInvulnerable(ExecutingMove@ attack, Pokemon@ target, bool& result)");
    SCRIPT_HOOK_FUNCTION(OnAttackMiss, "void OnAttackMiss(ExecutingMove@ attack, Pokemon@ target)");
    SCRIPT_HOOK_FUNCTION(ChangeAttackType,
                         "void ChangeAttackType(ExecutingMove@ attack, Pokemon@ target, uint8 hit, uint8& outType)");
    SCRIPT_HOOK_FUNCTION(
        ChangeEffectiveness,
        "void ChangeEffectiveness(ExecutingMove@ attack, Pokemon@ target, uint8 hit, float& effectiveness)");
    SCRIPT_HOOK_FUNCTION(
        PreventSecondaryEffects,
        "void PreventSecondaryEffects(ExecutingMove@ attack, Pokemon@ target, uint8 hit, bool& outResult)");
    SCRIPT_HOOK_FUNCTION(OnSecondaryEffect,
                         "void OnSecondaryEffect(ExecutingMove@ attack, Pokemon@ target, uint8 hit)");
    SCRIPT_HOOK_FUNCTION(OnAfterHits, "void OnAfterHits(ExecutingMove@ attack, Pokemon@ target)");
    SCRIPT_HOOK_FUNCTION(ModifyEffectChance,
                         "void ModifyEffectChance(ExecutingMove@ attack, Pokemon@ target, float& chance)");
    SCRIPT_HOOK_FUNCTION(ModifyIncomingEffectChance,
                         "void ModifyIncomingEffectChance(ExecutingMove@ attack, Pokemon@ target, float& chance)");

    SCRIPT_HOOK_FUNCTION(OverrideBasePower,
                         "void OverrideBasePower(ExecutingMove@ attack, Pokemon@ target, uint8 hit, uint8& chance)");
    SCRIPT_HOOK_FUNCTION(
        ChangeDamageStatsUser,
        "void ChangeDamageStatsUser(ExecutingMove@ attack, Pokemon@ target, uint8 hit, Pokemon@& user)");
    SCRIPT_HOOK_FUNCTION(BypassDefensiveStat,
                         "void BypassDefensiveStat(ExecutingMove@ attack, Pokemon@ target, uint8 hit, bool& bypass)");
    SCRIPT_HOOK_FUNCTION(BypassOffensiveStat,
                         "void BypassOffensiveStat(ExecutingMove@ attack, Pokemon@ target, uint8 hit, bool& bypass)");
    SCRIPT_HOOK_FUNCTION(ModifyStatModifier,
                         "void ModifyStatModifier(ExecutingMove@ attack, Pokemon@ target, uint8 hit, float& modifier)");
    SCRIPT_HOOK_FUNCTION(
        ModifyDamageModifier,
        "void ModifyDamageModifier(ExecutingMove@ attack, Pokemon@ target, uint8 hit, float& modifier)");
    SCRIPT_HOOK_FUNCTION(OverrideDamage,
                         "void OverrideDamage(ExecutingMove@ attack, Pokemon@ target, uint8 hit, int& damage)");

    SCRIPT_HOOK_FUNCTION(
        ModifyCriticalStage,
        "void ModifyCriticalStage(ExecutingMove@ attack, Pokemon@ target, uint8 hit, uint8& critStage)");
};

#undef SCRIPT_HOOK_FUNCTION
#endif // PKMNLIB_ANGELSCRIPTTYPEINFO_HPP
